#!/usr/bin/env python
# coding: utf-8

# In[27]:


import tkinter as tk
from tkinter.filedialog import asksaveasfile
from PIL import Image, ImageTk
import math

# Layout constants.
MARGIN = 5
WINDOW_WID = 500
WINDOW_HGT = 300

class ColorfulApp:
    # Create and manage the tkinter interface.
    def __init__(self, redraw_on_resize=False):
        self.redraw_on_resize = redraw_on_resize
        self.first_time = True
        self.image = None

        # Make the main interface.
        self.window = tk.Tk()
        self.window.title('ZoomImage')
        self.window.protocol('WM_DELETE_WINDOW', self.kill_callback)
        self.window.geometry(f'{WINDOW_WID}x{WINDOW_HGT}')

        # Build the rest of the UI.
        self.build_ui()

        # Set windowing bounds.
        self.set_window_bounds()

        # Display the window.
        self.window.focus_force()
        self.window.mainloop()

    def build_ui(self):
        # Menubar
        menubar = tk.Menu(self.window)
        self.window.config(menu=menubar)

        # File menu
        file_menu = tk.Menu(menubar, tearoff=False)
        file_menu.add_command(label='Save As...', accelerator="Ctrl+S", command=self.save_as)
        file_menu.add_separator()
        file_menu.add_command(label='Exit', command=self.kill_callback)
        menubar.add_cascade(label='File', menu=file_menu)

        # Scale menu
        scale_menu = tk.Menu(menubar, tearoff=False)
        scale_menu.add_command(label='Redraw', accelerator="F5", command=lambda: self.set_scale('redraw'))
        scale_menu.add_separator()
        scale_menu.add_command(label='x2', command=lambda: self.set_scale('2'))
        scale_menu.add_command(label='x4', command=lambda: self.set_scale('4'))
        scale_menu.add_command(label='x8', command=lambda: self.set_scale('8'))
        scale_menu.add_command(label='x16', command=lambda: self.set_scale('16'))
        scale_menu.add_command(label='Full Scale', command=lambda: self.set_scale('full scale'))
        menubar.add_cascade(label='Scale', menu=scale_menu)

        # Drawing canvas
        self.canvas = tk.Canvas(self.window, bg='black', borderwidth=1, highlightthickness=0, cursor='cross')
        self.canvas.pack(side=tk.TOP, padx=MARGIN, pady=MARGIN, expand=True, fill=tk.BOTH)

        # Initially we are not selecting an area.
        self.start_pos = None
        self.end_pos = None

        # Track Button-1 down, Button-1 motion, and Button-1 up.
        self.canvas.bind("<Button-1>", self.mouse_down)
        self.canvas.bind("<B1-Motion>", self.mouse_move)
        self.canvas.bind("<ButtonRelease-1>", self.mouse_up)

        # Watch for resize.
        self.canvas.bind("<Configure>", self.configured)

        # Redraw on F5.
        self.window.bind('<F5>', lambda event: self.set_scale('redraw'))

        # Save on Ctrl+S.
        self.window.bind('<Control-s>', lambda event: self.save_as())

    def set_window_bounds(self):
        '''Set the world coordinate bounds.'''
        # full_wXXX gives the full scale coordinates for the window.
        # desired_wXXX gives the desired world window coordinates.
        # wXXX gives the current world window coordinates.
        # vXXX gives the current viewport coordinates.
        self.full_wxmin = -1.05
        self.full_wxmax = 1.05
        self.full_wymin = -1.05
        self.full_wymax = 1.05
        self.min_width = 0.1
        self.min_height = 0.1
        self.desired_wxmin = self.full_wxmin
        self.desired_wxmax = self.full_wxmax
        self.desired_wymin = self.full_wymin
        self.desired_wymax = self.full_wymax

    def save_as(self):
        if self.image is None: return
        file = asksaveasfile(initialfile = 'Colorful.png',
            defaultextension='.png', filetypes=[('All Files', '*.*'), ('Image Files', '*.png *.jpg *.bmp')])
        if file is None: return
        self.image.save(file.name)

    def set_scale(self, scale_factor):
        '''The user has invoked a Scale menu item. Set the new world bounds and then redraw.'''
        # Get the desired world width and height.
        if scale_factor == 'redraw':
            # Just redraw. Leave the bounds unchanged.
            pass
        elif scale_factor == 'full scale':
            # Reset to full scale.
            self.desired_wxmin = self.full_wxmin
            self.desired_wxmax = self.full_wxmax
            self.desired_wymin = self.full_wymin
            self.desired_wymax = self.full_wymax
        else:
            # Scale by a number.
            scale = int(scale_factor)
            wid = (self.desired_wxmax - self.desired_wxmin) * scale
            hgt = (self.desired_wymax - self.desired_wymin) * scale
            cx = (self.desired_wxmax + self.desired_wxmin) / 2
            cy = (self.desired_wymax + self.desired_wymin) / 2
            self.desired_wxmin = cx - wid / 2
            self.desired_wxmax = self.desired_wxmin + wid
            self.desired_wymin = cy - hgt / 2
            self.desired_wymax = self.desired_wymin + hgt

        # If the width/height is larger than that of
        # full scale, use the full scale width/height.
        if self.desired_wxmax  - self.desired_wxmin > self.full_wxmax - self.full_wxmin:
            self.desired_wxmax = self.full_wxmax
            self.desired_wxmin = self.full_wxmin
        if self.desired_wymax  - self.desired_wymin > self.full_wymax - self.full_wymin:
            self.desired_wymax = self.full_wymax
            self.desired_wymin = self.full_wymin

        # Get the desired window bounds.
        wxmin = self.desired_wxmin
        wxmax = self.desired_wxmax
        wymin = self.desired_wymin
        wymax = self.desired_wymax

        # Get the width and height.
        w_wid = (wxmax - wxmin)
        w_hgt = (wymax - wymin)

        # Adjust to preserve aspect ratio.
        w_aspect = w_wid / w_hgt
        v_wid = self.vxmax - self.vxmin
        v_hgt = self.vymax - self.vymin
        v_aspect = v_wid / v_hgt
        if w_aspect > v_aspect:
            # World window bounds are too short and wide. Make taller.
            w_hgt = w_wid / v_aspect
        else:
            # World window bounds are too tall and thin. Make wider.
            w_wid = w_hgt * v_aspect

        # Set the new world window bounds.
        cx = (wxmin + wxmax) / 2
        cy = (wymin + wymax) / 2
        self.wxmin = cx - w_wid / 2
        self.wxmax = self.wxmin + w_wid
        self.wymin = cy - w_hgt / 2
        self.wymax = self.wymin + w_hgt

        # Redraw.
        self.draw()

    def mouse_down(self, event):
        self.start_pos = (event.x, event.y)
        self.end_pos = (event.x, event.y)

    def mouse_move(self, event):
        # Ignore extraneous events.
        if self.start_pos is None: return

        # Delete any previous selection rectangle.
        self.canvas.delete('selection_rectangle')

        # Create a new selection rectangle.
        self.end_pos = (event.x, event.y)
        self.canvas.create_rectangle(self.start_pos, self.end_pos,
            outline='red', tag='selection_rectangle')
        self.canvas.create_rectangle(
            self.start_pos, self.end_pos, dash=(1, 5),
            outline='yellow', tag='selection_rectangle')

    def mouse_up(self, event):
        # Ignore extraneous events.
        if self.start_pos is None: return
        self.canvas.delete('selection_rectangle')

        # Get the selected bounds.
        xmin = min(self.start_pos[0], self.end_pos[0])
        xmax = max(self.start_pos[0], self.end_pos[0])
        ymin = min(self.start_pos[1], self.end_pos[1])
        ymax = max(self.start_pos[1], self.end_pos[1])
        self.start_pos = None
        self.end_pos = None
        
        # Do nothing if the selected area is empty.
        if xmin == xmax or ymin == ymax: return

        # Call area_selected to redraw.
        self.area_selected(xmin, xmax, ymin, ymax)

    def area_selected(self, xmin, xmax, ymin, ymax):
        # Convert into world coordinates.
        wxmin, wymin = self.d_to_w(xmin, ymin)
        wxmax, wymax = self.d_to_w(xmax, ymax)

        # Make sure the width and height aren't too small.
        if wxmax - wxmin < self.min_width:
            cx = (wxmax + wxmin) / 2
            wxmin = cx - self.min_width / 2
            wxmax = wxmin + self.min_width
        if wymax - wymin < self.min_height:
            cy = (wymax + wymin) / 2
            wymin = cy - self.min_height / 2
            wymax = wymin + self.min_height

        # Set the desired world coordinates.
        self.desired_wxmin = wxmin
        self.desired_wxmax = wxmax
        self.desired_wymin = wymin
        self.desired_wymax = wymax

        # Redraw.
        self.set_scale('redraw')

    def configured(self, event):
        '''The window may have been resized. Reset the viewport coordinates and redraw.'''
        # Do nothing unless the window is in a normal or zoomed state.
        if self.window.state() != 'normal' and self.window.state() != 'zoomed': return

        self.vxmin = 0
        self.vxmax = self.canvas.winfo_width() - 1
        self.vymin = 0
        self.vymax = self.canvas.winfo_height() - 1

        # Redraw.
        if self.first_time or self.redraw_on_resize:
            self.set_scale('redraw')
            self.first_time = False

    def kill_callback(self):
        self.window.destroy()

    def draw(self):
        '''Draw an image.'''
        # Set scale factors for drawing.
        self.x_scale = (self.vxmax - self.vxmin) / (self.wxmax - self.wxmin)
        self.y_scale = (self.vymax - self.vymin) / (self.wymax - self.wymin)

        # Make an image to fit the viewport.
        mode = 'RGB'
        image_wid = self.vxmax - self.vxmin
        image_hgt = self.vymax - self.vymin
        bg_color = (0, 0, 0)
        self.image = Image.new(mode, (image_wid, image_hgt), bg_color)

        # Access the image's pixels.
        self.pixels = self.image.load()

        # Get the viewport's bounds in world coordinates.
        w_vxmin, w_vymin = self.d_to_w(0, 0)
        w_vxmax, w_vymax = self.d_to_w(image_wid - 1, image_hgt - 1)

        # Loop over the pixels.
        wwid = self.full_wxmax - self.full_wxmin
        whgt = self.full_wymax - self.full_wymin
        wcx = 0
        wcy = 0
        for x in range(image_wid):
            for y in range(image_hgt):
                # Convert (x, y) from viewport coordinates to world coordinates.
                wx, wy = self.d_to_w(x, y)

                # Calculate this point's color.
                # If it's not within -1 <= x, y <= 1, make it black.
                if wx < self.full_wxmin or wx > self.full_wxmax or \
                   wy < self.full_wymin or wy > self.full_wymax:
                        color = (0, 0, 0)
                else:
                    # Distance from center.
                    dx = wx - wcx
                    dy = wy - wcy
                    distance = math.sqrt(dx * dx + dy * dy)
                    angle = math.atan2(dy, dx)
                    r = int(255 * math.sin(angle))
                    g = int(255 * distance % 50)
                    b = int(255 * math.cos(angle))
                    color = (r, g, b)

                # Set the pixel's color.
                self.pixels[x, y] = color

        # Convert the PIL image into an ImageTk.PhotoImage.
        self.photo_image = ImageTk.PhotoImage(self.image)

        # Display the result.
        self.canvas.delete('all')
        self.display_image = self.canvas.create_image(0, 0, anchor=tk.NW, image=self.photo_image)

    def w_to_d(self, x, y):
        '''Convert a point from world to device coordinates.'''
        vx = self.vxmin + (x - self.wxmin) * self.x_scale
        vy = self.vymin + (y - self.wymin) * self.y_scale
        return vx, vy

    def d_to_w(self, x, y):
        '''Convert a point from device to world coordinates.'''
        vx = self.wxmin + (x - self.vxmin) / self.x_scale
        vy = self.wymin + (y - self.vymin) / self.y_scale
        return vx, vy


# In[28]:


# Do not allow auto-redraw on resize.
ColorfulApp(False)
print('Done')


# In[ ]:




